/**
  *  \file client/dialogs/resetdlg.cc
  *  \brief Reset Location dialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/resetdlg.h"
#include "client/wtile.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "game/reset.h"
#include "ui/widgets/button.h"
#include "util/strtab.h"
#include "ui/widgets/richstatic.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/richparser.h"
#include "client/widgets/objlist.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"

namespace {
    /** Reset options tile.
        Displays the options. */
    class WResetOptionsTile : public WTile {
     public:
        WResetOptionsTile(GReset& reset);
        void getLayoutInfo(LayoutInfo& info);
        void drawData(GfxCanvas& can);

     private:
        void init();
        void computeLayout();

        GReset& reset;
        ObjectHolder h;
        int left_x, right_x, right_wi;
    };

    /** Reset dialog.
        Displays an introductory text, options, and buttons. */
    class WResetDialog : public UIWindow {
     public:
        WResetDialog(GReset& reset);
        bool handleEvent(const UIEvent& event, bool second_pass);
        void update();
        void doList();

     private:
        GReset& reset;
        WResetOptionsTile* tile;
        UIButton* btn_ok;
        ObjectHolder h;

        void init();
    };
}

static const char*const left_side[] = {
    N_("Cargo, structures, builds"),
    N_("Missions, orders, names"),
};

static const char*const right_side[] = {
    N_("no"),
    N_("yes"),
};

static const char keys[] = { 'C', 'M' };

static const char HELP_PAGE[] = "pcc2:reset";

/*************************** WResetOptionsTile ***************************/

WResetOptionsTile::WResetOptionsTile(GReset& reset)
    : WTile(GfxPoint(1, 1)),
      reset(reset)
{
    init();
}

void
WResetOptionsTile::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(right_x + 5, countof(left_side) * font_heights[FONT_TITLE]);
    info.flags = 0;
}

void
WResetOptionsTile::drawData(GfxCanvas& can)
{
    const int h = font_heights[FONT_TITLE];
    const int y = h/2;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);

    /* Left side */
    ctx.setTextAlign(0, 1);
    ctx.setColor(UIColor::tc_Static);
    for (uint_t i = 0; i < countof(left_side); ++i)
        outText(ctx, left_x, y + h*i, _(left_side[i]));

    /* Right side */
    ctx.setTextAlign(2, 1);
    ctx.setColor(UIColor::tc_Green);
    ctx.setSolidBackground();
    outTextF(ctx, getExtent().w - 5, y,     right_wi, _(right_side[(reset.getMode() & GReset::ResetCargo) != 0]));
    outTextF(ctx, getExtent().w - 5, y+h,   right_wi, _(right_side[(reset.getMode() & GReset::ResetMissions) != 0]));
}

void
WResetOptionsTile::init()
{
    using std::tolower;
    computeLayout();
    setLayout(LeftJustified, 0, false);
    const int d = font_heights[FONT_TITLE];
    for (uint_t i = 0; i < countof(keys); ++i) {
        addControl(GfxRect(0, d*i+1, d-2, d-2), h.add(new UIButton(string_t(&keys[i], 1), tolower(keys[i]))));
    }
}

void
WResetOptionsTile::computeLayout()
{
    int lw = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(left_side, countof(left_side)));
    int rw = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(right_side, countof(right_side)));

    left_x   = font_heights[FONT_TITLE] + 5;
    right_x  = left_x + lw + rw;
    right_wi = rw;
}


/****************************** WResetDialog *****************************/

WResetDialog::WResetDialog(GReset& reset)
    : UIWindow(_("Reset Location"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      reset(reset),
      tile(0),
      btn_ok(0),
      h()
{
    init();
}

void
WResetDialog::init()
{
    // VBox
    //   UIRichStatic
    //   WResetOptionsTile
    //   HBox
    //     "OK"
    //     "Cancel"
    //     "List"
    //     UISpacer
    //     "Help"
    tile = &h.add(new WResetOptionsTile(reset));
    btn_ok = &h.add(new UIButton(_("OK"), SDLK_RETURN));

    setLayout(UIVBoxLayout::instance5);
    add(h.add(new UIRichStatic(RichParser::textFromXml(_("This will reset all units at this place to their state at the beginning of the turn. "
                                                         "Please choose which areas you want to reset. "
                                                         "<b>Cargo</b> will reset everything that interacts with cargo (buildings, transfers, etc.), "
                                                         "<b>Missions</b> will reset everything else.")),
                               font_ems[FONT_NORMAL]*30)));
    add(*tile);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(*btn_ok);
    g.add(h.add(new UIButton(_("Cancel"), SDLK_ESCAPE)));
    g.add(h.add(new UIButton(_("L - List units"), 'l')));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(_("Help"), 'h')));
    add(g);
    add(h.add(new WHelpWidget(HELP_PAGE)));
    add(h.add(new UIQuit(0)));
    pack();
    update();
}

bool
WResetDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'm':
            reset.setMode(reset.getMode() ^ GReset::ResetMissions);
            update();
            return true;

         case 'c':
            reset.setMode(reset.getMode() ^ GReset::ResetCargo);
            update();
            return true;

         case 'l':
            doList();
            return true;

         case SDLK_RETURN:
            if (reset.isReady()) {
                stop(1);
            }
            return true;

         case SDLK_ESCAPE:
            stop(0);
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WResetDialog::update()
{
    tile->drawWidget();
    btn_ok->setState(st_Disabled, !reset.isReady());
}

void
WResetDialog::doList()
{
    // FIXME: can we get rid of the const_cast?
    WObjectList(0, 10, 20*font_ems[FONT_NORMAL],
                const_cast<GObjectList&>(reset.getObjects())).doStandardDialog2(_("Affected Units"), HELP_PAGE);
}

/** "Reset" dialog.
    Displays a dialog to configure the given GReset instance.
    \retval true User confirmed, call reset.commit()
    \retval false User cancelled */
bool
doResetDialog(GReset& reset)
{
    if (reset.getObjects().size() > 0) {
        WResetDialog dlg(reset);
        ui_root->centerWidget(dlg);
        return ui_root->runModal(dlg) != 0;
    } else {
        return false;
    }
}
