/**
  *  \file client/dialogs/searchdlg.cc
  *  \brief Search Dialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "client/dialogs/searchdlg.h"
#include "client/dialogs/gadialog.h"
#include "client/dialogs/selmgr.h"
#include "client/lru.h"
#include "client/screens.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/lrupopup.h"
#include "client/widgets/objlist.h"
#include "client/widgets/searchobj.h"
#include "client/widgets/selswitch.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/sel.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "client/screens/minescreen.h"
#include "client/widgets/decaymsg.h"
#include "util/misc.h"
#include "client/screens/ionscreen.h"
#include "client/screens/ufoscreen.h"
#include "ui/widgets/quit.h"

/** Global instance of search query. This stores the last search query. */
TUserSearchQuery search_query;

namespace {
    class WSearchDialog : public UIWindow {
        ObjectHolder           h;
        GObjectList&           result;
        WSearchObjectSelector* obj_selector;
        WSearchObjectSelector* obj_selector2;
        UIStandardListbox*     type_selector;
        UIInputLine*           query_input;
        UIButton*              query_dropdown;
        WObjectList*           result_list;
        UIStaticText*          result_info;
        StringList             type_list;
        bool                   with_global;

        GObject*               current_obj;

        bool prepareQuery(TParsedSearchQuery& q);
        void doMarkResult(bool value);

     public:
        WSearchDialog(GObjectList& result, GObject* current_obj, bool with_global);
        ~WSearchDialog();

        void init(bool with_global);
        void loadSearchQuery();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        bool handleEvent(const UIEvent& event, bool second_pass);

        void onResultChange();
        void onSearch();
        void onGoTo();
        void onGoToChart();
        void onGoToShip(int screen);
        void onGoToPlanet(int screen);
        void onQueryDropdown();

        void doSelectionCommands(UIBaseWidget& btn);

        GObjectReference getCurrentObject() const;
        bool preferBases() const;
    };

    enum {
        cm_GoTo         = 2001,
        cm_Selection    = 2003,
        cm_Global       = 2004
    };

    enum {
        stopCancel,
        stopOK,
        stopEvent,
        stopGlobal
    };

    class WSearchUserInterfaceBinding : public IntUserInterfaceChildBinding {
     public:
        WSearchUserInterfaceBinding(WSearchDialog& dlg)
            : event(),
              dialog(dlg)
            { }
        virtual IntValue* getProperty(IntUserInterfaceProperty p)
            {
                return parent.getProperty(p);
            }
        virtual void setProperty(IntUserInterfaceProperty p, IntValue* v)
            {
                parent.setProperty(p, v);
            }
        virtual void postEvent(std::auto_ptr<IntUserInterfaceEvent> event)
            {
                switch (event->command) {
                 case iucPopupConsole:
                 case iucExitRace:
                 case iucExitRaceNoSave:
                 case iucExitClient:
                 case iucChartSetView:
                 case iucGoToScreen:
                 case iucGoToChart:
                    // Repost
                    this->event = event;
                    dialog.stop(stopEvent);
                    break;

                 case iucOverlayMessage:
                    // Execute
                    showDecayingMessage(IntValue::toStringOf(event->args[0], false));
                    break;

                 case iucGoToMinefield:
                    // Execute
                    doMinefieldScreen();
                    break;

                 case iucGoToUfo:
                    doUfoScreen();
                    break;

                 case iucGoToStorm:
                    doIonStormScreen();
                    break;

                 case iucUseKeymap:
                    // Ignore?
                    break;

                 case iucExitDialog:
                    dialog.stop(stopOK);
                    break;
                }
            }
        virtual void enumContexts(ptr_vector<IntContext>& ctx)
            {
                parent.enumContexts(ctx);
            }
        void repostEvent()
            {
                parent.postEvent(event);
            }
     private:
        std::auto_ptr<IntUserInterfaceEvent> event;
        WSearchDialog& dialog;
    };
}

/** Constructor. Builds the complete search dialog. */
WSearchDialog::WSearchDialog(GObjectList& result, GObject* current_obj, bool with_global)
    : UIWindow(_("Search Object"), BLUE_WINDOW, COLOR_GRAY),
      result(result),
      with_global(with_global),
      current_obj(current_obj)
{
    init(with_global);
}

/** Destructor. */
WSearchDialog::~WSearchDialog()
{ }

bool
WSearchDialog::prepareQuery(TParsedSearchQuery& query)
{
    search_query.search_kind = TUserSearchQuery::Match(type_selector->getKey());
    search_query.search_obj  = obj_selector->getObjects() | obj_selector2->getObjects();
    search_query.query       = query_input->getText();

    return prepareSearchQuery(query, search_query);
}

/** Mark/unmark search result.
    \param value true to mark, false to unmark */
void
WSearchDialog::doMarkResult(bool value)
{
    for (GObjectList::iterator i = result.begin(); i != result.end(); ++i) {
        if (i->isGameObject()) {
            (*i)->setIsMarked(value);
        }
    }
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
}

void
WSearchDialog::init(bool with_global)
{
    /* Initialize types */
    type_list.add(0, _("Name/Id/Comment"));
    type_list.add(1, _("Expression true"));
    type_list.add(2, _("Expression false"));
    type_list.add(3, _("Location"));

    // VBox
    //   HBox 1
    //     VBox 11
    //       Static "Search objects:"
    //       UIFrameGroup: WSearchObjectSelector
    //     VBox 12
    //       Static "Search type:"
    //       UIFrameGroup: UIStringList
    //       UISpacer
    //   HBox 2
    //     UIInputLine
    //     UIButton "dropdown"
    //   HBox 3
    //     UIStaticText(?) status.......
    //     UISpacer
    //     UIButton "Go"
    //   UIFrameGroup [HBox] 4
    //     UIListbox (result)
    //     UIScrollbar
    //   HBox 5
    //     UIButton "Go to"
    //     UIButton "Close"
    //     UIButton "Mark"
    //     UIButton "Global"
    //     UISpacer
    //     UIButton "Help"
    setLayout(UIVBoxLayout::instance5);

    // Group 1
    UIGroup& group_1        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group_11       = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group_12       = h.add(new UIGroup(UIVBoxLayout::instance5));

    obj_selector = &h.add(new WSearchObjectSelector(1000, 0, 5));
    group_11.add(h.add(new UIStaticText(_("Search Objects:"), UIStaticText::StaticTitle)));
    group_11.add(UIFrameGroup::wrap(h, *obj_selector));
    group_11.add(h.add(new UISpacer()));

    type_selector = &h.add(new UIStandardListbox(1001,
                                                 type_list.getCount(),
                                                 getTextWidth(*fonts[FONT_NORMAL], type_list),
                                                 type_list));
    group_12.add(h.add(new UIStaticText(_("Search Type:"), UIStaticText::StaticTitle)));
    group_12.add(UIFrameGroup::wrap(h, *type_selector));
    obj_selector2 = &h.add(new WSearchObjectSelector(1002, 5, 6));
    group_12.add(UIFrameGroup::wrap(h, *obj_selector2));

    group_1.add(group_11);
    group_1.add(group_12);
    add(group_1);

    // Group 2
    UIGroup& group_2 = h.add(new UIGroup(UIHBoxLayout::instance0));
    query_input = &h.add(new UIInputLine(INT_MAX, 20, 0, UIInputLine::ilf_Framed, 1003));
    group_2.add(*query_input);
    query_dropdown = &h.add(new UIButton(0, UTF_DOWN_ARROW, 0));
    query_dropdown->sig_fired.add(this, &WSearchDialog::onQueryDropdown);
    group_2.add(*query_dropdown);
    add(group_2);

    // Group 3
    UIGroup& group_3 = h.add(new UIGroup(UIHBoxLayout::instance5));
    result_info = &h.add(new UIStaticText(string_t(), UIStaticText::StaticTitle));
    group_3.add(*result_info);
    group_3.add(h.add(new UISpacer()));
    group_3.add(h.add(new UIButton(0, _("Search!"), 0)).addCall(this, &WSearchDialog::onSearch));
    add(group_3);

    // Group 4
    UIFrameGroup& group_4 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    result_list = &h.add(new WObjectList(1004, 12, 20*font_ems[FONT_NORMAL], result));
    result_list->sig_dblclick.add(this, &WSearchDialog::onGoTo);
    group_4.add(*result_list);
    group_4.add(h.add(new UIScrollbar(*result_list)));
    add(group_4);

    // Group 5
    UIGroup& group_5 = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_5.add(h.add(new UIButton(cm_GoTo,      _("Go to"),     0))           .addCommand());
    group_5.add(h.add(new UIButton(              _("Close"),     SDLK_ESCAPE)));
    group_5.add(h.add(new UIButton(cm_Selection, _("Mark..."),   ss_Alt + 'm')).addCommand());
    if (with_global)
        group_5.add(h.add(new UIButton(cm_Global, _("Global..."), ss_Alt + 'g')).addCommand());
    group_5.add(h.add(new UISpacer()));
    group_5.add(h.add(new UIButton(_("Help"), ss_Alt + 'h')));
    add(group_5);

    // Administrivia
    add(h.add(new UIFocusIterator(1000, 1004, UIFocusIterator::fi_Tab)));
    add(h.add(new WSelectionSwitcher()));
    add(h.add(new WHelpWidget("pcc2:search")));
    add(h.add(new UIQuit(stopCancel)));

    pack();

    // Initialize dialog
    loadSearchQuery();
    query_input->setFocusedWidget();

    onResultChange();
}

void
WSearchDialog::loadSearchQuery()
{
    obj_selector->setObjects(search_query.search_obj);
    obj_selector2->setObjects(search_query.search_obj);
    type_selector->scrollToKey(search_query.search_kind);
    query_input->setText(search_query.query);
}

bool
WSearchDialog::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case cm_GoTo:
        onGoTo();
        return true;
     case cm_Selection:
        doSelectionCommands(sender);
        return true;
     case cm_Global:
        stop(stopGlobal);
        return true;
     default:
        return UIWindow::handleCommand(cmd, sender);
    }
}

bool
WSearchDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        if (event.key.code == SDLK_DOWN && query_input->hasState(st_Focused)) {
            onQueryDropdown();
            return true;
        }
        if (event.key.code == SDLK_F7) {
            query_input->setFocusedWidget();
            return true;
        }
        if (event.key.code == SDLK_F1 && result_list->hasState(st_Focused)) {
            onGoToShip(1);
            return true;
        }
        if (event.key.code == SDLK_F2 && result_list->hasState(st_Focused)) {
            onGoToPlanet(2);
            return true;
        }
        if (event.key.code == SDLK_F3 && result_list->hasState(st_Focused)) {
            onGoToPlanet(3);
            return true;
        }
        if ((event.key.code == SDLK_F4 || event.key.code == SDLK_RETURN + ss_Shift) && result_list->hasState(st_Focused)) {
            onGoToChart();
            return true;
        }
        if (event.key.code == SDLK_F6 && result_list->hasState(st_Focused)) {
            onGoToShip(6);
            return true;
        }
        if (event.key.code == SDLK_RETURN) {
            if (result_list->hasState(st_Focused)) {
                onGoTo();
            } else {
                onSearch();
            }
            return true;
        }
        if (event.key.code == SDLK_ESCAPE) {
            stop(stopCancel);
            return true;
        }
        if (event.key.code == ss_Alt + '.') {
            bool search = doSelectionManagerFromSearch();
            getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
            if (search) {
                loadSearchQuery();
                onSearch();
            }
            return true;
        }
        if (event.key.code == ss_Alt + 'o') {
            obj_selector2->setFocusedWidget();
            obj_selector2->onItemClick(0, 0);
            return true;
        }
    }

    return UIWindow::handleEvent(event, second_pass);
}

void
WSearchDialog::onResultChange()
{
    bool blank = result.size() == 0;
    result_list->setState(st_Disabled, blank);
    setStateOf(cm_GoTo,      st_Disabled, blank);
    setStateOf(cm_Selection, st_Disabled, blank);
    setStateOf(cm_Global,    st_Disabled, blank);

    if (blank) {
        result_info->setText(string_t());
    } else {
        result_info->setText(format(_("%d result%!1{s%}"), numToString(result.size())));
    }

    /* Perform a null size change to request relayout. */
    result_info->getParent()->onResize(result_info->getParent()->getExtent(), false);

    /* onResize will not redraw the result_info widget correctly in case it
       reduced its size, because it is transparent and nobody feels like
       redrawing the background as a reaction to a null size change. We
       therefore request the whole window to be redrawn. Since the result
       list has changed, most of it will be redrawn anyway. */
    drawWidget();
}

void
WSearchDialog::onSearch()
{
    // Do we have some objects?
    if (obj_selector->getObjects() == 0) {
        messageBox(_("Please select some object types to search."), _("Search Object"));
        obj_selector->setFocusedWidget();
        return;
    }

    // Prepare query
    TParsedSearchQuery parsed_query;
    search_query.search_kind = TUserSearchQuery::Match(type_selector->getKey());
    search_query.search_obj  = obj_selector->getObjects() | obj_selector2->getObjects();
    search_query.query       = query_input->getText();
    if (!prepareSearchQuery(parsed_query, search_query)) {
        return;
    }

    // Clear result list
    result.clear();
    TSearchResultStatus status;

    // Execute query
    executeSearchQuery(parsed_query, status, result);
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    // Update result
    result_list->handleListChange(GObjectReference());
    onResultChange();

    // Tell user why we didn't find anything
    if (result.size() == 0) {
        if (!status.had_object) {
            // No object at all
            messageBox(_("There are no objects of the requested kind."), _("Search Object"));
            obj_selector->setFocusedWidget();
        } else if (status.had_error && !status.did_evaluate && status.last_error.size()) {
            // We didn't evaluate the expression successfully, all invocations failed.
            // This means there's something wrong with the expression.
            messageBox(format(_("Error in search query: %s"), status.last_error), _("Search Object"));
            query_input->setFocusedWidget();
        } else {
            // Nothing found
            messageBox(_("Your query didn't match any object."), _("Search Object"));
            query_input->setFocusedWidget();
        }
    } else {
        // If the current object is among the results, mark it
        result_list->setFocusedWidget();
        if (current_obj) {
            int index = result.getIndexFor(*current_obj);
            if (index > 0) {
                result_list->scrollTo(index-1, 1);
            }
        }
    }

    // If query was successful, is an expression query, not a match-any (i.e. empty),
    // and not predefined, put it on LRU list
    if ((!status.had_error || status.did_evaluate)
        && parsed_query.match != TParsedSearchQuery::MatchAny
        && (search_query.search_kind == TUserSearchQuery::MatchTrue || search_query.search_kind == TUserSearchQuery::MatchFalse)
        && predef_search.findValue(search_query.query) >= predef_search.size())
    {
        lru_search.pushNewLimit(new LRUListItem(search_query.query,
                                                (search_query.search_kind == TUserSearchQuery::MatchTrue ? "[]" : "[!]") + search_query.query));
    }
}

void
WSearchDialog::onGoTo()
{
    if (getCurrentObject().isGameObject()) {
        if (with_global) {
            postGoToObject(IntUserInterfaceBinding::get(), *getCurrentObject(), preferBases());
        } else {
            stop(stopOK);
        }
    }
}

void
WSearchDialog::onGoToChart()
{
    if (getCurrentObject().isGameObject()) {
        // Only if not called from global actions which cannot handle IntUserInterfaceBinding operations
        if (with_global) {
            // We normally would have to lock this against objects that are not on the map,
            // but let's assume that invisible objects are not found.
            if (GMapObject* mo = dynamic_cast<GMapObject*>(&*getCurrentObject())) {
                postGoToChart(IntUserInterfaceBinding::get(), mo->getPos(), mo);
            }
        }
    }
}

void
WSearchDialog::onGoToShip(int screen)
{
    if (getCurrentObject().isGameObject()) {
        if (with_global) {
            if (GShip* sh = dynamic_cast<GShip*>(&*getCurrentObject())) {
                postGoToScreen(IntUserInterfaceBinding::get(), screen, sh->getId());
            }
        }
    }
}

void
WSearchDialog::onGoToPlanet(int screen)
{
    if (getCurrentObject().isGameObject()) {
        if (with_global) {
            if (GPlanet* pl = dynamic_cast<GPlanet*>(&*getCurrentObject())) {
                postGoToScreen(IntUserInterfaceBinding::get(), screen, pl->getId());
            }
        }
    }
}

void
WSearchDialog::onQueryDropdown()
{
    query_input->setFocusedWidget();

    const LRUListItem* item = WLRUPopup()
        .withList(_("Last queries"), lru_search)
        .withList(_("Predefined queries"), predef_search)
        .doPopup(query_input->getText(), *query_dropdown);

    if (item != 0) {
        // User has selected an item. Parse it.
        string_t value = item->getValue();
        uint16_t type = TUserSearchQuery::MatchTrue;
        uint16_t obj = 0;
        if (value.size() != 0 && value[0] == '[') {
            string_t::size_type i = 1;
            while (i < value.size() && value[i] != ']') {
                switch (value[i]) {
                 case 'S': case 's':
                    obj |= TUserSearchQuery::SearchShips;
                    break;
                 case 'P': case 'p':
                    obj |= TUserSearchQuery::SearchPlanets;
                    break;
                 case 'B': case 'b':
                    obj |= TUserSearchQuery::SearchBases;
                    break;
                 case '!':
                    type = TUserSearchQuery::MatchFalse;
                    break;
                }
                ++i;
            }
            value = strLTrim(value.substr(i+1));
        } else {
            obj = TUserSearchQuery::AllObjects;
        }

        // If it specifies an object type, set that
        if (obj != 0) {
            obj_selector->setObjects(obj);
        }

        // Set type and query
        type_selector->scrollTo(type, 1);
        query_input->setText(value);
    }
}

void
WSearchDialog::doSelectionCommands(UIBaseWidget& btn)
{
    enum { Mark, MarkOnly, Unmark };

    StringList list;
    list.add(Mark,     _("Mark found objects"));
    list.add(MarkOnly, _("Mark only found objects"));
    list.add(Unmark,   _("Unmark found objects"));

    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    if (!lb.doStandardMenu(btn))
        return;

    switch (lb.getKey()) {
     case Mark:
        doMarkResult(true);
        break;
     case MarkOnly:
        GSelection().copyTo(getDisplayedTurn().getCurrentUniverse());
        doMarkResult(true);
        break;
     case Unmark:
        doMarkResult(false);
        break;
    }
}

GObjectReference
WSearchDialog::getCurrentObject() const
{
    return result_list->getCurrentObject();
}

inline bool
WSearchDialog::preferBases() const
{
    return (obj_selector->getObjects() & TUserSearchQuery::SearchPlanets) == 0;
}

/** Perform search (F7) function. */
void
doSearch(bool immediate, GObject* current_obj)
{
    GObjectList result;
    WSearchDialog dialog(result, current_obj, true);
    WSearchUserInterfaceBinding binding(dialog);
    ui_root->centerWidget(dialog);
    ui_root->add(dialog);
    if (immediate) {
        dialog.onSearch();
    }
    int r = ui_root->run();
    ui_root->close(dialog);
    if (r == stopEvent) {
        binding.repostEvent();
    } else if (r == stopGlobal) {
        doGlobalActions(result);
    } else {
        // nothing
    }
}

/** Perform search (F7) function with specified result list. */
void
doSearch(GObjectList& list)
{
    WSearchDialog dialog(list, 0, false);
    ui_root->centerWidget(dialog);
    ui_root->add(dialog);
    if (list.size() != 0) {
        /* If the list is nonempty, assume it is the result of a
           previous search, so rebuild it. */
        dialog.onSearch();
    }
    ui_root->run();
    ui_root->close(dialog);
}

/** Initialize search function. This discards the previous search query. */
void
initSearch()
{
    search_query.search_kind = TUserSearchQuery::MatchName;
    search_query.search_obj  = TUserSearchQuery::AllObjects;
    search_query.query       = "";
}

const ModuleInitializer mod_searchdlg[] = {
    { OnEnterNoGame | OnEnterDirectory | OnEnterGame, initSearch },
    { 0, 0 },
};
