/**
  *  \file client/lrupredef.cc
  *  \brief Least-Recently-Used Lists - Predefined Lists
  *
  *  This module defines the global LRUList instances used in the PCC2
  *  GUI. It also provides loading of predefined item lists from the
  *  "expr.cc"/"expr.usr" file through a modinit hook.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/lru.h"
#include "cpluslib/format.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/filepars.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/pref.h"
#include "io/dirs.h"
#include "arch/platform.h"
#include "util/except.h"

static const char LRU_FILE[] = "lru.ini";

/**************************** Predefined Lists ***************************/

LRUList predef_shiplabels, predef_planetlabels, predef_search;
LRUList lru_shiplabels, lru_planetlabels, lru_search;

static const struct {
    const char* name;
    LRUList* p[2];
} lru_list_names[] = {
    { "SHIPLABELS",   { &predef_shiplabels,   &lru_shiplabels } },
    { "PLANETLABELS", { &predef_planetlabels, &lru_planetlabels } },
    { "FIND",         { &predef_search,       &lru_search } }
};

static LRUList*
getListByName(const string_t& str, bool lru)
{
    for (uint_t i = 0; i < countof(lru_list_names); ++i) {
        if (str == lru_list_names[i].name) {
            return lru_list_names[i].p[lru];
        }
    }
    return 0;
}

namespace {
    class ListFileParser : public FileParser {
        LRUList* section;
        bool lru;
     public:
        explicit ListFileParser(bool lru);
        void syntaxError(int line_id);
        void process(int line_id, string_t line);
    };
}

ListFileParser::ListFileParser(bool lru)
    : FileParser(";#"),
      section(0),
      lru(lru)
{
    setUtf8();
}

void
ListFileParser::syntaxError(int line_id)
{
    console.write(LOG_WARN, format(_("%s:%d: file format error -- line ignored"), getFileName(), line_id));
}

void
ListFileParser::process(int line_id, string_t line)
{
    string_t::size_type n = line.find_first_not_of(" \t");
    if (n == string_t::npos) {
        // cannot happen
    } else if (line[n] == '[') {
        // section
        string_t::size_type p = line.find(']', n);
        if (p == string_t::npos) {
            syntaxError(line_id);
        } else {
            section = getListByName(strUCase(string_t(line, n+1, p-n-1)), lru);
        }
    } else if (section != 0) {
        // line in section
        string_t::size_type p = line.find("  ", n);
        if (p == string_t::npos) {
            syntaxError(line_id);
        } else {
            section->pushNew(new LRUListItem(string_t(line, n, p-n), strTrim(string_t(line, p+1))));
        }
    } else {
        // unknown section
    }
}

/********************************* Saving ********************************/

static void
saveLRUList(TextFile& tf, const char* name, LRUList& list)
{
    tf.writeLine(format("[%s]", name));
    for (uint32_t index = 0, n = list.size(); index < n; ++index) {
        const LRUListItem& it = list[index];

        // Sanitize the name so we'll be able to re-parse it
        string_t name = it.getName();
        string_t::size_type i = 0;
        bool deleting = true;
        while (i < name.size()) {
            uint8_t ch = name[i];
            if ((deleting && ch <= ' ') || ch < ' ') {
                name.erase(i, 1);
            } else {
                ++i;
            }
            deleting = (ch <= ' ');
        }

        // Save
        tf.write(name);
        tf.write("  ");
        tf.writeLine(it.getValue());
    }
    tf.writeLine();
}

/***************************** Initialisation ****************************/

static void
loadPredefinedLists()
{
    predef_shiplabels.clear();
    predef_planetlabels.clear();
    predef_search.clear();

    if (Ptr<Stream> s = Stream::openFileNT(Directory::makeFileName(Platform::getUserSettingsDirectory(1), "expr.ini"), Stream::C_READ)) {
        ListFileParser(false).parseFile(*s);
    }
    ListFileParser(false).parseGameFile("expr.cc");
    ListFileParser(false).parseGameFile("expr.usr");
}

static void
loadLRULists()
{
    if (Ptr<Stream> s = Stream::openFileNT(Directory::makeFileName(Platform::getUserSettingsDirectory(1), LRU_FILE), Stream::C_READ)) {
        ListFileParser(true).parseFile(*s);
    }
}

static void
saveLRULists()
{
    string_t fileName = Directory::makeFileName(Platform::getUserSettingsDirectory(1), LRU_FILE);
    try {
        // Create preferences directory
        createUserPreferencesDirectory();

        // Create file
        Ptr<Stream> s = Stream::openFile(fileName, Stream::C_CREATE);
        TextFile tf(*s);
        tf.setUtf8();
        for (uint_t i = 0; i < countof(lru_list_names); ++i) {
            saveLRUList(tf, lru_list_names[i].name, *lru_list_names[i].p[true]);
        }
        tf.sync();
    }
    catch (FileProblemException& fpe) {
        console.write(LOG_WARN, format("%s: %s", fpe.getFileName(), fpe.what()));
    }
}

const ModuleInitializer mod_lru[] = {
    { OnEnterNoGame | OnEnterDirectory, loadPredefinedLists },
    { OnStartup, loadLRULists },
    { OnLeftDirectory | OnShutdown, saveLRULists },
    { 0, 0 },
};
