/**
  *  \file client/widgets/ftrinfo.cc
  *  \brief WFighterInfo widget
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/widgets/ftrinfo.h"
#include "cpluslib/format.h"
#include "game/host.h"
#include "game/player.h"
#include "game/racename.h"
#include "game/spec.h"
#include "game/team.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "game/config.h"

static string_t formatCost(GCost c)
{
    string_t result;
    static const GCargoType order[] = { el_Money, el_Supplies, el_Tritanium, el_Duranium, el_Molybdenum };
    const char* labels[] = { _("mc"), _("sup"), "T", "D", "M" };
    for (size_t i = 0; i < countof(order); ++i) {
        int32_t n = c.get(order[i]);
        if (n != 0) {
            if (!result.empty()) {
                result += ", ";
            }
            result += numToString(n);
            result += " ";
            result += labels[i];
            for (size_t j = i+1; j < countof(order); ++j) {
                if (c.get(order[j]) == n) {
                    c.set(order[j], 0);
                    result += "/";
                    result += labels[j];
                }
            }
        }
    }
    if (result.empty()) {
        result = "-";
    }
    return result;
}

WFighterInfo::WFighterInfo()
    : player_id(::getPlayerId())
{ }

WFighterInfo::~WFighterInfo()
{ }

void
WFighterInfo::drawContent(GfxCanvas& can)
{
    // Modelled after the torpedo view, because it might overlay it:
    //   Type:    fighter
    //   Kill:
    //   Destroy:
    //   Recharge: (recharge time in seconds)
    //   Strikes:
    //   Fighter Cost:

    /* Recharge time for PHost: Actual recharge rate is BayRechargeRate + nbays*BayRechargeBonus.
       We cannot model the bonus, but we know that the actual rate includes minimum one bonus.
       We can add a relation sign to help the user.

       Given a rate, the average recharge per tick is (rate-1)/2, making the average time
       2000/(rate-1).

       Recharge time for THost: a fighter is launched if rand(1..20) <= nbays. That is, a one-bay
       ship launches at a rate of 1/20, a two-bay ship at 2/20, a three-bay ship at 3/20, etc.,
       making the per-bay launch rate 1/20, and the recharge time 20. */

    /* Strikes for PHost: this is explicit in the StrikesPerFighter. We do not model
       the speeds here; they can become significant, but usually aren't.

       Strikes for THost: fighters fire in [-19,+19] around the enemy, a 39 ly range.
       Fighters move at speed 4, units at speed 1, giving a difference speed of 4 or 5.
       This allows for 7 to 10 strikes. */

    const GFighter ftr(getFighterForPlayer(player_id));
    const int freeFighters = config.FreeFighters(player_id);

    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getExtent(), FONT_NORMAL);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Kill:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Destroy:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Recharge:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Strikes:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Fighter Cost:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\xE2\x80\xA2 Auto-build:"));

    tw.newRColumn(3*font_ems[FONT_NORMAL]);
    GfxPoint typeMark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());  // leave 'type' cell blank for now
    tw.writeVariableCell(UIColor::tc_Green, numToString(ftr.getKill()));
    tw.writeVariableCell(UIColor::tc_Green, numToString(ftr.getDamage()));
    if (host.isTHost()) {
        tw.writeVariableCell(UIColor::tc_Green, "20");
        tw.writeVariableCell(UIColor::tc_Green, "7" UTF_EN_DASH "10");
    } else {
        /* Note that positive bonus means "<=" relation! */
        const int bonus = config.BayRechargeBonus(player_id);
        const int rate = config.BayRechargeRate(player_id) + bonus;
        const char*const prefix = bonus < 0 ? UTF_GEQ " " : bonus > 0 ? UTF_LEQ " " : "";
        if (rate > 1) {
            tw.writeVariableCell(UIColor::tc_Green, prefix + numToString(std::max(config.BayLaunchInterval(player_id), 2000 / (rate+1))));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, "-");
        }
        tw.writeVariableCell(UIColor::tc_Green, numToString(config.StrikesPerFighter(player_id)));
    }
    GfxPoint priceMark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    GfxPoint freeMark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Green, string_t());
    tw.writeFixedCell(UIColor::tc_Green, string_t());
    tw.writeFixedCell(UIColor::tc_Green, string_t());
    tw.writeFixedCell(UIColor::tc_Green, "s");
    tw.writeFixedCell(UIColor::tc_Green, string_t());
    tw.writeFixedCell(UIColor::tc_Green, string_t());
    tw.writeFixedCell(UIColor::tc_Green, string_t());

    tw.finish();

    ctx.setSolidBackground();
    ctx.setTextAlign(0, 0);
    ctx.setColor(UIColor::tc_Green);
    ctx.useFont(*fonts[FONT_NORMAL]);
    outTextF(ctx, typeMark.x, typeMark.y, getExtent().x + getExtent().w - typeMark.x,
             format(_("%s Fighter"), player_racenames.getAdjName(player_id)));
    outTextF(ctx, priceMark.x, priceMark.y, getExtent().x + getExtent().w - priceMark.x, formatCost(ftr.getCost()));

    outTextF(ctx, freeMark.x, freeMark.y, getExtent().x + getExtent().w - freeMark.x,
             freeFighters > 0
             ? string_t(format(_("%d per turn for %s each"), freeFighters, formatCost(config.FreeFighterCost(player_id))))
             : string_t(_("none")));

    can.update(getExtent());
}

void
WFighterInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = font_sizes[FONT_NORMAL].scaledBy(28, 7);
}

void
WFighterInfo::setPlayerId(int n)
{
    if (n != player_id) {
        player_id = n;
        drawWidget();
    }
}
