/**
  *  \file game/utilmsg.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/utilmsg.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/hullfunc.h"
#include "game/racename.h"
#include "game/simplembox.h"
#include "game/tables.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/misc.h"
#include "util/nls.h"

GUtilMessageParser::GUtilMessageParser()
{ }

GUtilMessageParser::~GUtilMessageParser()
{ }

void
GUtilMessageParser::loadDefinition(Stream& in)
{
    // ex utilmsgs.pas:InitUtilViewer
    TextFile tf(in);
    string_t line;
    Description* current = 0;

    while (tf.getLine(line)) {
        line = strTrim(line);
        if (line.empty() || line[0] == ';') {
            continue;
        }
        string_t::size_type p = line.find_first_of(",=");
        if (p == string_t::npos) {
            console.write(LOG_ERROR, format(_("%s:%d: missing delimiter"), in.getName(), tf.getLineNr()));
        } else if (line[p] == '=') {
            // It's an assignment
            if (!current) {
                continue;
            }

            string_t lhs = strTrim(line.substr(0, p));
            string_t rhs = strTrim(line.substr(p+1));
            bool ok = true;
            if (stringMatch("Heading", lhs)) {
                current->header = rhs;
            } else if (stringMatch("Text", lhs)) {
                current->text += rhs;
                current->text += '\n';
            } else if (stringMatch("Alias", lhs)) {
                ok = stringValue(rhs, current->alias);
            } else if (stringMatch("Loop", lhs)) {
                ok = stringValue(rhs, current->loop_size);
            } else if (stringMatch("Fixed", lhs)) {
                ok = stringValue(rhs, current->fixed_size);
            } else if (stringMatch("Max", lhs)) {
                ok = stringValue(rhs, current->load_limit);
            }

            if (!ok) {
                console.write(LOG_ERROR, format(_("%s:%d: invalid number"), in.getName(), tf.getLineNr()));
            }
        } else {
            // New record
            int id;
            if (!stringValue(strTrim(line.substr(0, p)), id)) {
                console.write(LOG_ERROR, format(_("%s:%d: invalid number"), in.getName(), tf.getLineNr()));
                current = 0;
            } else {
                current = new Description();
                util_descs.push_back(current);
                current->id = id;
                current->alias = -1;
                current->load_limit = -1;
                current->loop_size = 0;
                current->fixed_size = 0;
                current->name = strTrim(line.substr(p+1));
                current->header = "(-h0000)";
            }
        }
    }
}

void
GUtilMessageParser::loadFile(Stream& in, GSimpleMailbox& out)
{
    class Reader : public GUtilReader {
     public:
        Reader(GSimpleMailbox& out, GUtilMessageParser& parent)
            : out(out), parent(parent)
            { }

        virtual bool process(int rec_id, std::size_t size, const char data[])
            {
                parent.addRecord(out, rec_id, size, data);
                return true;
            }
        virtual void failure(Stream& /*s*/)
            { }
     private:
        GSimpleMailbox& out;
        GUtilMessageParser& parent;
    };

    Reader(out, *this).read(in);
}

const GUtilMessageParser::Description*
GUtilMessageParser::getDescriptionById(int id) const
{
    // ex utilmsgs.pas:GetRecordById
    for (size_t i = 0, n = util_descs.size(); i < n; ++i) {
        if (util_descs[i]->id == id) {
            return util_descs[i];
        }
    }
    return 0;
}

void
GUtilMessageParser::addRecord(GSimpleMailbox& out, int rec_id, std::size_t size, const char data[])
{
    // ex CUtilMailbox.Msg2Edit, CUtilMailbox.ReadFile
    const Description* desc = getDescriptionById(rec_id);
    if (desc == 0) {
        out.addMessage(_("(-h0000)<<< Unknown >>>"), format(_("Record type %d is unknown."), rec_id));
    } else {
        // Save header
        string_t header = desc->header;
        string_t name = desc->name;

        // Chase 'alias' links
        int paranoia = 10;
        while (desc != 0 && desc->alias >= 0 && paranoia > 0) {
            desc = getDescriptionById(desc->alias);
            --paranoia;
        }
        if (desc == 0) {
            out.addMessage(format("%s<<< %s >>>", decodeRecord(data, size, header), name),
                           string_t(format(_("Record type %d, %d bytes"), rec_id, size)) + "\n\n" + _("Unknown reference target in record definition."));
        } else {
            // Apply load_limit
            size_t orig_size = size;
            if (desc->load_limit > 0) {
                size = std::min(size, size_t(desc->load_limit));
            }

            // Apply loop
            if (desc->loop_size == 0) {
                out.addMessage(format("%s<<< %s >>>", decodeRecord(data, size, header), name),
                               string_t(format(_("Record type %d, %d bytes"), rec_id, orig_size)) + "\n\n" + decodeRecord(data, size, desc->text));
            } else {
                int part_nr = 0;
                for (size_t i = size_t(desc->fixed_size); i < size; i += size_t(desc->loop_size)) {
                    std::vector<char> tmp;
                    tmp.reserve(desc->fixed_size + desc->loop_size);
                    tmp.insert(tmp.end(), &data[0], &data[desc->fixed_size]);
                    tmp.insert(tmp.end(), &data[i], &data[std::min(i + size_t(desc->loop_size), size)]);
                    ++part_nr;
                    out.addMessage(format("%s<<< %s >>>", decodeRecord(&tmp[0], tmp.size(), header), name),
                                   string_t(format(_("Record type %d, part %d"), rec_id, part_nr)) + "\n\n" + decodeRecord(&tmp[0], tmp.size(), desc->text));
                }
            }
        }
    }
}

string_t
GUtilMessageParser::decodeRecord(const char data[], size_t size, const string_t& tpl)
{
    // ex utilmsgs.pas:DecodeRecord
    size_t idx = 0;
    const size_t len = tpl.size();

    bool had_undef = false;
    string_t out_line;
    string_t out_text;
    size_t cur_dat = 0;
    while (idx < len) {
        char c = tpl[idx++];
        if (c == '%') {
            // Determine cursor and process flags
            size_t i = 0;
            bool tmp = false;
            bool ignore_undef = false;
            bool force_print = false;
            int32_t L = -1;

            while (idx < len) {
                c = tpl[idx++];
                if (c >= '0' && c <= '9') {
                    i = 10*i + (c - '0');
                    tmp = true;
                } else if (c == '?') {
                    ignore_undef = true;
                } else if (c == '!') {
                    force_print = true;
                } else {
                    break;
                }
            }
            if (tmp) {
                cur_dat = i;
            }

            // Handle item
            // If we left the above loop by idx >= len (syntax error), this will interpret the
            // '%' or final flag as type code. Why not.
            string_t app;
            switch (c) {
             case 'S':
             {
                 size_t slen = 0;
                 for (int i = 0; i < 2; ++i) {
                     if (idx < len && tpl[idx] >= '0' && tpl[idx] <= '9') {
                         slen = 10*slen + tpl[idx++] - '0';
                     }
                 }
                 cur_dat += slen;
                 if (cur_dat >= slen && cur_dat <= size) {
                     app = getBasicStringN(&data[cur_dat-slen], slen);
                 } else {
                     if (!ignore_undef) {
                         had_undef = true;
                     }
                 }
                 break;
             }
             case 'X':
             case 'l':
             case 'F':
                // long
                cur_dat += 4;
                if (cur_dat >= 4 && cur_dat <= size) {
                    L = getInt32(&data[cur_dat-4]);
                } else {
                    L = -1;
                }
                goto FormatNum;
             case 'b':
                // byte
                ++cur_dat;
                if (cur_dat >= 1 && cur_dat <= size) {
                    L = uint8_t(data[cur_dat-1]);
                } else {
                    L = -1;
                }
                goto FormatNum;
             case '%':
                app = "%";
                break;
             case '|':
                app = "  ";
                break;
             default:
                // words
                cur_dat += 2;
                if (cur_dat >= 2 && cur_dat <= size) {
                    L = getInt16(&data[cur_dat-2]);
                } else {
                    L = -1;
                }
             FormatNum:
                if (!force_print && ((L == -1) || ((L == 0 && std::strchr("prhg", c) != 0)))) {
                    if (!ignore_undef) {
                        had_undef = true;
                    }
                } else {
                    if (c == 'g' && L >= 0 && L <= 9) {
                        app = getNativeGovernmentName(L);
                    } else if (c == 'h' && L > 0 && L <= getNumHulls()) {
                        app = getHull(L).getName();
                    } else if (c == 'H') {
                        // { This is a little more technical than GetBasicFunctionTitle, which
                        //   would report a "Alchemy: 1 Mineral -> 1 Fuel" device, but I think
                        //   it fits better in here }
                        if (const GBasicHullFunction* hf = GHullFunctionData::instance.getBasicFunctionById(L)) {
                            app = hf->getName();
                        } else {
                            app = format("<%d>", L);
                        }
                    } else if (c == 'n' && L >= 0 && L <= 9) {
                        app = getNativeRaceName(L);
                    } else if (c == 'B') {
                        for (int i = 0; i <= 15; ++i) {
                            if ((L & (1L << i)) != 0) {
                                app += format("%d ", i);
                            }
                        }
                        if (app.empty()) {
                            app = _("none");
                        } else {
                            app.erase(app.size()-1);
                        }
                    } else if (c == 'p') {
                        if (haveDisplayedTurn()) {
                            GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
                            if (univ.isValidPlanetId(L)) {
                                app = univ.getPlanet(L).getName(GObject::PlainName);
                            }
                        }
                    } else if (c == 'r' && L > 0 && L <= NUM_OWNERS) {
                        app = player_racenames.getShortName(L);
                    } else if (c == 'u') {
                        if (const Description* def = getDescriptionById(L)) {
                            app = def->name;
                        } else {
                            app = format("%d", L);
                        }
                    } else if (c == 'W') {
                        app = format("%04d", L);
                    } else if (c == 'R') {
                        app = format("%5d", L);
                    } else if (c == 'X') {
                        app = format("%08X", uint32_t(L));
                    } else if (c == 'x') {
                        app = format("%04X", uint16_t(L & 65535));
                    } else if (c == 'F') {
                        app = format("%04d", std::abs(L));
                        app.insert(app.size()-3, ".");
                        if (L < 0) {
                            app.insert(0, "-");
                        }
                    } else if (c == '(') {
                        int32_t cnt = 0;
                        while (idx < len) {
                            c = tpl[idx++];
                            if (c == ',') {
                                ++cnt;
                            } else if (c == ')') {
                                break;
                            } else if (c == '\n') {
                                // skip
                            } else {
                                if (cnt == L) {
                                    app += c;
                                }
                            }
                        }
                        if (app.empty()) {
                            app = format("%d", L);
                        }
                    } else {
                        app = format("%d", L);
                    }
                }
            }
            out_line += app;
        } else if (c == '\n') {
            // New line
            if (!had_undef && (!out_line.empty() || out_text.size() <= 2 || out_text[out_text.size()-2] != '\n')) {
                out_text += out_line;
                out_text += '\n';
            }
            out_line.clear();
            had_undef = false;
        } else {
            // normal character
            out_line += c;
        }
    }

    // Add potential partial line
    out_text += out_line;

    return out_text;
}
