/**
  *  \file gfx/gfxutil.cc
  *  \brief Graphics Utilities
  *
  *  Utility classes used for implementing the graphics and UI system.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <ostream>
#include <algorithm>
#include "cpluslib/bitops.h"
#include "gfx/gfx.h"
#include "gfx/gfxutil.h"

static const uint8 solid_init[]   = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
static const uint8 gray50_init[]  = { 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA };
static const uint8 gray25_init[]  = { 0xAA, 0x00, 0x55, 0x00, 0xAA, 0x00, 0x55, 0x00 };
static const uint8 ltslash_init[] = { 0x88, 0x44, 0x22, 0x11, 0x88, 0x44, 0x22, 0x11 };

/** Fill pattern: solid. */
const GfxFillPattern SOLID_FILL(solid_init);
/** Fill pattern: 50% gray. */
const GfxFillPattern GRAY50_FILL(gray50_init);
/** Fill pattern: 25% gray. */
const GfxFillPattern GRAY25_FILL(gray25_init);
/** Fill pattern: 50% gray, alternative version. This is the inverse
    of GRAY50_FILL. */
const GfxFillPattern GRAY50_FILL2(gray50_init+1);
/** Fill pattern: slashes. */
const GfxFillPattern LTSLASH_FILL(ltslash_init);

/** Construct blank pattern. Constructs a pattern that is all-zero. */
GfxFillPattern::GfxFillPattern() throw()
{
    std::fill(pat, pat+8, uint8(0));
}

/** Construct pattern from byte array.
    \param init pointer to 8 bytes */
GfxFillPattern::GfxFillPattern(const uint8 init[]) throw()
{
    std::copy(init, init+8, pat);
}

/** Construct pattern with specified value. Constructs a pattern that
    contains the specified value in each line. */
GfxFillPattern::GfxFillPattern(uint8 init) throw()
{
    std::fill(pat, pat+8, init);
}

/** Check for blank pattern. \returns true iff the pattern is all-zero,
    i.e. when it is used for filling something, nothing would happen. */
bool
GfxFillPattern::isBlank() const throw()
{
    for (int i = 0; i < 8; ++i)
        if (pat[i])
            return false;
    return true;
}

/** Check for black pattern. \returns true iff the pattern is all-one,
    i.e. filling something with it would result in a completely filled
    surface. */
bool
GfxFillPattern::isBlack() const throw()
{
    for (int i = 0; i < 8; ++i)
        if (pat[i] != 255)
            return false;
    return true;
}

/** Shift pattern to the left. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftLeft(int amount) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] = rol8(pat[i], amount);
    return *this;
}

/** Shift pattern to the right. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftRight(int amount) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] = ror8(pat[i], amount);
    return *this;
}

/** Shift pattern up. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftUp(int amount) throw()
{
    GfxFillPattern tmp = *this;
    for (int i = 0; i < 8; ++i)
        pat[i] = tmp.pat[unsigned(i+amount) % 8];
    return *this;
}

/** Shift pattern down. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftDown(int amount) throw()
{
    GfxFillPattern tmp = *this;
    for (int i = 0; i < 8; ++i)
        pat[i] = tmp.pat[unsigned(i-amount) % 8];
    return *this;
}

/** `Or' two patterns. The result has 1-bits where either of the two
    patterns had 1-bits. \returns reference to this pattern to allow
    chaining. */
GfxFillPattern&
GfxFillPattern::operator |= (const GfxFillPattern& rhs) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] |= rhs.pat[i];
    return *this;
}

/** `Or' pattern with value. \overload */
GfxFillPattern&
GfxFillPattern::operator |= (uint8 value) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] |= value;
    return *this;
}

/** `And' two patterns. The result has 1-bits where both of the two
    patterns had 1-bits. \returns reference to this pattern to allow
    chaining. */
GfxFillPattern&
GfxFillPattern::operator &= (const GfxFillPattern& rhs) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] &= rhs.pat[i];
    return *this;
}

/** `And' pattern and value. \overload */
GfxFillPattern&
GfxFillPattern::operator &= (uint8 value) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] &= value;
    return *this;
}

/** `Xor' two patterns. The result has 1-bits where either of the two
    patterns (but not both) had 1-bits. \returns reference to this
    pattern to allow chaining. */
GfxFillPattern&
GfxFillPattern::operator ^= (const GfxFillPattern& rhs) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] ^= rhs.pat[i];
    return *this;
}

/** `Xor' pattern and value. \overload */
GfxFillPattern&
GfxFillPattern::operator ^= (uint8 value) throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] ^= value;
    return *this;
}

/** Invert pattern. Turns black into white and vice versa.
    \returns reference to this pattern to allow chaining */
GfxFillPattern&
GfxFillPattern::invert() throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] = ~ pat[i];
    return *this;
}

/** Flip pattern horizontally. Swaps left and right.
    \returns reference to this pattern to allow chaining */
GfxFillPattern&
GfxFillPattern::flipHorizontal() throw()
{
    for (int i = 0; i < 8; ++i)
        pat[i] = flipBits8(pat[i]);
    return *this;
}

/** Flip pattern vertically. Swaps top and bottom.
    \returns reference to this pattern to allow chaining */
GfxFillPattern&
GfxFillPattern::flipVertical() throw()
{
    std::swap(pat[0], pat[7]);
    std::swap(pat[1], pat[6]);
    std::swap(pat[2], pat[5]);
    std::swap(pat[3], pat[4]);
    return *this;
}


//////////////////////////////////////////////////////////////////////

/** Construct bounding rectangle for GfxCanvas.
    Constructs a rectangle that covers the whole area of \c can. */
GfxRect::GfxRect(GfxCanvas& can)
{
    GfxPoint pt = can.getSize();
    x = y = 0;
    w = pt.x;
    h = pt.y;
}

/** Clip at size. Modify this rectangle so that it lies entirely
    within \c GfxRect(0, 0, cw, ch). This is slightly more efficient
    than clipRect(). */
void
GfxRect::clipSize(int cw, int ch) throw()
{
    register long sw = w, sh = h;
    if (x < 0)
        sw += x, x = 0;
    if (y < 0)
        sh += y, y = 0;
    if (x + sw > cw)
        sw = cw - x;
    if (y + sh > ch)
        sh = ch - y;
    w = (sw < 0) ? 0 : sw;
    h = (sh < 0) ? 0 : sh;
}

/** Clip at rectangle. Modify this rectangle so that it lies entirely
    within the other rectangle, \c r.

    This function is also available under the alternate name intersect().

    \post result.contains(p) iff orig.contains(p) && r.contains(p) for
    all points p */
void
GfxRect::clipRect(const SDL_Rect& r) throw()
{
    register long sw = w, sh = h;
    if (x < r.x)
        sw += x - r.x, x = r.x;
    if (y < r.y)
        sh += y - r.y, y = r.y;
    if (x + sw > r.w + r.x)
        sw = r.w - x + r.x;
    if (y + sh > r.h + r.y)
        sh = r.h - y + r.y;
    w = (sw < 0) ? 0 : sw;
    h = (sh < 0) ? 0 : sh;
}

/** Include rectangle. Computes the smallest possible rectangle which
    includes this one as well r, and modifies this rectangle to
    contain these bounds.

    \post r.contains(p) || orig.contains(p) => result.contains(p) for
    all points p (the converse obviously does not hold). */
void
GfxRect::include(const SDL_Rect& r) throw()
{
    if (r.w <= 0 || r.h <= 0)
        return;
    if (!exists()) {
        *this = r;
        return;
    }

    if (r.x < x) {
        w += x - r.x;
        x = r.x;
    }
    if (r.y < y) {
        h += y - r.y;
        y = r.y;
    }
    if (r.x + r.w - x > w)
        w = r.x + r.w - x;
    if (r.y + r.h - y > h)
        h = r.y + r.h - y;
}

/** Include point. Computes the smallest possible rectangle which
    includes this one as well as pt, and modifies this rectangle to
    contain these bounds.

    \post r.contains(p) => result.contains(p) for all points p
    \post result.contains(pt) */
void
GfxRect::include(GfxPoint pt) throw()
{
    include(GfxRect(pt, 1, 1));
}

/** Compare rectangles for equality.
    \return true iff rhs completely identical to *this
    \note empty rectangles do not compare equal unless their
    x,y is also identical */
bool
GfxRect::operator==(const SDL_Rect& rhs) const throw()
{
    return x == rhs.x
        && y == rhs.y
        && w == rhs.w
        && h == rhs.h;
}

/** Compare rectangles for inequality.
    \return true iff rhs differs somehow from *this
    \note empty rectangles do not compare equal unless their
    x,y is also identical */
bool
GfxRect::operator!=(const SDL_Rect& rhs) const throw()
{
    return x != rhs.x
        || y != rhs.y
        || w != rhs.w
        || h != rhs.h;
}

/** Move to point without changing the size. Moves the rectangle to
    point /where/ without touching the size.
    \param where new top-left corner for this rectangle
    \return relative movement done. E.g. if a rectangle located at
    (5,5) is moved to (20,30), this returns (15,25).
    \post x == where.x && y == where.y */
GfxPoint
GfxRect::moveTo(GfxPoint where) throw()
{
    GfxPoint moved_by (where.x - x, where.y - y);
    moveBy(moved_by);
    return moved_by;
}

/** Check whether this rectangle completely contains r.
    \param r rectangle
    \return true iff r.contains(x,y) => contains(x,y) for all x,y. */
bool
GfxRect::contains(const GfxRect& r) const throw()
{
    if (!exists())
        return false;
    if (!r.exists())
        return true;

    /* check whether r AND *this == r */
    GfxRect rcopy = r;
    rcopy.intersect(*this);
    return rcopy == r;
}

/** Move this rectangle such that it is contained within \c other. 
    Does not change our size. */
void
GfxRect::moveInto(const GfxRect& other) throw()
{
    if (x < other.x)
        x = other.x;
    if (y < other.y)
        y = other.y;

    int avail_x = other.x + other.w - x;
    if (avail_x < w)
        x = other.x + other.w - w;

    int avail_y = other.y + other.h - y;
    if (avail_y < h)
        y = other.y + other.h - h;
}

/** Center this rectangle within \c other. */
void
GfxRect::centerWithin(const GfxRect& other) throw()
{
    moveToEdge(other, 1, 1, 0);
}

/** Move this rectangle to edge of \c other.
    \param other  Bounding rectangle
    \param x_pos  Relative X position (0: left, 1: center, 2: right)
    \param y_pos  Relative Y position (0: top, 1: center, 2: bottom)
    \param offset Distance to edge. When anchored at an edge (0 or 2),
                  leave that many pixels from that edge. */
void
GfxRect::moveToEdge(const GfxRect& other, int x_pos, int y_pos, int offset) throw()
{
    const int virt_w = int(other.w) - w;
    const int virt_h = int(other.h) - h;

    x = other.x + x_pos * virt_w / 2 - offset * (x_pos-1);
    y = other.y + y_pos * virt_h / 2 - offset * (y_pos-1);
}

/** Reduce this rectangle's width from the left.
    Removes \c pix pixels from the left (less if the rectangle is narrower).
    Same as splitX, but does not return a result.
    \param pix number of pixels to remove */
void
GfxRect::consumeX(int pix)
{
    if (pix < 0) {
        pix = 0;
    }
    if (pix > w) {
        pix = w;
    }
    w -= pix;
    x += pix;
}

void
GfxRect::consumeRightX(int pix)
{
    w -= std::min(int(w), std::max(0, pix));
}

/** Split rectangle vertically.
    Removes \c pix pixels from the left (less if the rectangle is narrower), like consumeX(int).
    Returns the removed rectangle.
    \param pix number of pixels to remove */
GfxRect
GfxRect::splitX(int pix)
{
    if (pix < 0) {
        pix = 0;
    }
    if (pix > w) {
        pix = w;
    }
    const int resultX = x;
    w -= pix;
    x += pix;
    return GfxRect(resultX, y, pix, h);
}

/** Reduce this rectangle's height from the top.
    Removes \c pix pixels from the top (less if the rectangle is shorter).
    Same as splitY, but does not return a result.
    \param pix number of pixels to remove */
void
GfxRect::consumeY(int pix)
{
    if (pix < 0) {
        pix = 0;
    }
    if (pix > h) {
        pix = h;
    }
    h -= pix;
    y += pix;
}

/** Split rectangle horizontally.
    Removes \c pix pixels from the top (less if the rectangle is shorter), like consumeY(int).
    Returns the removed rectangle.
    \param pix number of pixels to remove */
GfxRect
GfxRect::splitY(int pix)
{
    if (pix < 0) {
        pix = 0;
    }
    if (pix > h) {
        pix = h;
    }
    const int resultY = y;
    h -= pix;
    y += pix;
    return GfxRect(x, resultY, w, pix);
}


//////////////////////////////////////////////////////////////////////

/*! \class GfxRectSet
    \brief Rectangle Set

    Objects of this class contain an arbitrary set of rectangles. The
    class provides the usual set operations. This class is, among
    others, used in redraw management.

    \invariant this class contains a list of rectangle which are all
    disjoint. The disjointness is managed by all manipulator methods. */

/** Construct empty set.
    \post contains(x,y) == false for any x,y */
GfxRectSet::GfxRectSet()
{ }

/** Construct rectangular set. The set will contain the one rectangle r. */
GfxRectSet::GfxRectSet(const GfxRect& r)
{
    the_list.push_back(r);
}

/** Intersect this rectangle set with r.
    \post r.contains(x,y)==false => contains(x,y)==false
    \todo maybe this is more efficient to do in-place? */
void
GfxRectSet::intersect(const GfxRect& r)
{
    std::list<GfxRect> l2;
    for(std::list<GfxRect>::iterator i = the_list.begin(); i != the_list.end(); ++i) {
        GfxRect q(*i);
        q.intersect(r);
        if (q.exists ())
            l2.push_back(q);
    }
    the_list.swap(l2);
}

/** Rectangle Set Subtraction. Constructs in \c l2 the set of rectangles
    in \c other which are outside \c r. Rectangles are split when needed. */
void
GfxRectSet::doSubtract(const GfxRect& r, std::list<GfxRect>& l2,
                       std::list<GfxRect>& other)
{
    for(std::list<GfxRect>::iterator i = other.begin(); i != other.end(); ++i) {
        GfxRect q(*i);
        q.intersect(r);
        if (!q.exists()) {
            l2.push_back(*i);
        } else {
            int qby = q.y + q.h, qbx = q.x + q.w;
            int iby = i->y + i->h, ibx = i->x + i->w;
            if(q.y > i->y)
                l2.push_back(GfxRect(i->x, i->y, i->w, q.y - i->y));
            if(q.x > i->x)
                l2.push_back(GfxRect(i->x, q.y, q.x - i->x, q.h));
            if(qbx < ibx)
                l2.push_back(GfxRect(qbx, q.y, ibx - qbx, q.h));
            if(qby < iby)
                l2.push_back(GfxRect(i->x, qby, i->w, iby - qby));
        }
    }
}

/** Add rectangle to set. This does not necessarily add a copy of r to
    the set; if this set already contains parts of r, only the remaining
    parts are added.
    \param r rectangle
    \post r.contains(x,y) => contains(x,y)
    \post contains(r) */
void
GfxRectSet::add(const GfxRect& r)
{
    std::list<GfxRect> l2;
    l2.push_back(r);
    for(iterator i = the_list.begin(); i != the_list.end(); ++i) {
        std::list<GfxRect> result;
        doSubtract(*i, result, l2);
        l2.swap(result);
    }

    the_list.splice(the_list.end(), l2);
}

/** Remove rectangle from set.
    \param r rectangle
    \post r.contains(x,y) => !contains(x,y) */
void
GfxRectSet::remove(const GfxRect& r)
{
    std::list<GfxRect> l2;
    doSubtract(r, l2, the_list);
    the_list.swap(l2);
}

/** Check if point is in set.
    \param x,y coordinates
    \returns true iff (x,y) is in this set. */
bool
GfxRectSet::contains(int x, int y) const
{
    for(iterator i = begin(); i != end(); ++i)
        if(i->contains(x, y))
            return true;
    return false;
}

/** Check if rectangle in set.
    \param r rectangle
    \return true iff r is completely contained in *this. */
bool
GfxRectSet::contains(const GfxRect& r) const
{
    std::list<GfxRect> l2;
    l2.push_back(r);
    for(iterator i = the_list.begin(); i != the_list.end(); ++i) {
        std::list<GfxRect> result;
        doSubtract(*i, result, l2);
        l2.swap(result);
    }

    return !l2.empty();
}

/** Compute bounding rectangle.
    \returns the smallest possible rectangle which entirely
    contains this rectangle set */
GfxRect
GfxRectSet::getBoundingRectangle() const
{
    GfxRect br;
    for (iterator i = begin(); i != end(); ++i)
        br.include(*i);
    return br;
}

//////////////////////////////////////////////////////////////////////

std::ostream&
operator<<(std::ostream& os, const GfxRect& r)
{
    return os << "(" << r.x << ", " << r.y
              << ") +(" << r.w << ", " << r.h << ")";
}

std::ostream&
operator<<(std::ostream& os, GfxPoint p)
{
    return os << "(" << p.x << ", " << p.y << ")";
}
