/**
  *  \file int/vmfile.cc
  *  \brief Interpreter: Standard Virtual Machine File
  *
  *  This contains the necessary glue to produce VM files in users'
  *  game directories, using the engine from int/vmio.cc.
  *
  *  Format of VM file:
  *  <code>
  *      +0   4 BYTEs   Signature 'CCvm'
  *      +4  18 BYTEs   Timestamp
  *     +22     BYTE    Signature 26 (^Z)
  *     +23     BYTE    Format version (currently 100, PCC 1.x has 0 here)
  *     +24     WORD    Number of bytes following in header (currently 4)
  *     +26     WORD    Turn number
  *     +28     WORD    Player Id
  *  </code>
  *
  *  \todo Turn number validation is missing. PCC 1.x contains logic to
  *  - not load a VM file that has a higher turn number than the current
  *    turn
  *  - not save a VM file if the on-disk file has a higher turn number
  *    than the current turn
  *  This effectively means that scripts/auto tasks are ignored when
  *  reviewing a previous turn.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "int/vmfile.h"
#include "game/global.h"
#include "game/timestamp.h"
#include "int/exec.h"
#include "int/process.h"
#include "int/vmio.h"
#include "io/dirs.h"
#include "io/io.h"
#include "util/except.h"
#include "util/nls.h"

// VM file magic, 'CCvm'
static const uint32_t MAGIC = 16777216*'m' + 65536*'v' + 257*'C';

// Item offsets
enum {
    oSignature  = 0,
    oTimestamp  = 4,
    oEOF        = 22,
    oVersion    = 23,
    oHeaderSize = 24,
    oTurn       = 26,
    oPlayer     = 28
};

// Current version
enum {
    VERSION = 100
};

namespace {
    /** Determine whether we want to keep a loaded process.
        We want to terminate auto-tasks refering to nonexistant objects,
        which means either they do not have an associated object,
        or that object is not playable in the given category. */
    bool wantKeepProcess(const IntExecutionContext& p)
    {
        // Note that the conditions are different here than in c2ng,
        // because we run this immediately after creating the processes,
        // at which time they are runnable, not suspended.
        if (p.getState() != IntExecutionContext::psRunnable || !haveDisplayedTurn()) {
            // Strange state, better keep it
            return true;
        } else {
            const GObject* obj = p.getInvokingObject();
            GUniverse& univ = getCurrentTurn().getCurrentUniverse();
            switch (p.getProcessKind()) {
             case IntExecutionContext::pkShipTask:
                return obj != 0 && univ.ty_played_ships.isValidIndex(obj->getId());
             case IntExecutionContext::pkPlanetTask:
                return obj != 0 && univ.ty_played_planets.isValidIndex(obj->getId());
             case IntExecutionContext::pkBaseTask:
                return obj != 0 && univ.ty_played_bases.isValidIndex(obj->getId());
             case IntExecutionContext::pkDefault:
                break;
            }

            // Not an auto-task, keep it
            return true;
        }
    }

    /** Terminate unusuable auto tasks.
        Tasks are terminated by injecting a call to CC$AutoTerminate. */
    void terminateUnusableAutoTasks()
    {
        // Function to inject
        IntBCORef code = new IntBytecodeObject();
        code->setIsProcedure(true);
        code->addInstruction(IntOpcode::maPush, IntOpcode::sNamedVariable, code->addName("CC$AUTOTERMINATE"));
        code->addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 0);

        // Do it
        const IntProcessList& pl = getProcessList();
        for (IntProcessList::const_iterator i = pl.begin(); i != pl.end(); ++i) {
            if (*i != 0 && !wantKeepProcess(**i)) {
                IntExecutionFrame* f = new IntExecutionFrame(code);
                f->want_result = false;
                (*i)->pushNewFrame(f);
            }
        }

        runRunnableProcesses();
        killTerminatedProcesses();
    }
}

/** Load VM file.
    \param dir Directory to load from
    \param fn File name
    \param player Player number. Pass 0 to not validate the player number.
    \throw FileProblemException when there is a problem */
void
loadVM(Directory& dir, string_t fn, int player)
{
    Ptr<Stream> pvm = dir.openFileNT(fn, Stream::C_READ);
    if (pvm != 0) {
        // Load and validate header
        char header[30];
        pvm->readT(header, sizeof(header));
        uint16_t hdrSize = getUint16(&header[oHeaderSize]);
        if (getUint32(&header[oSignature]) != MAGIC
            || header[oVersion] != VERSION
            || hdrSize < 4
            || (player != 0 && getUint16(&header[oPlayer]) != player))
        {
            throw FileFormatException(*pvm, _("Invalid file header"));
        }

        // Move to data position
        pvm->seek(hdrSize + oTurn);

        // Load content
        IntVMLoadContext().load(*pvm, true);
        terminateUnusableAutoTasks();
        runRunnableProcesses();
        killTerminatedProcesses();
    }
}

/** Save VM file. When the file would be empty, it is erased.
    \param dir Directory to save to
    \param fn File name
    \param player Player number
    \throw FileProblemException when there is a problem */
void
saveVM(Directory& dir, string_t fn, int player)
{
    const IntProcessList& pl = getProcessList();
    if (pl.empty()) {
        // No processes to save, so erase the file
        dir.eraseNT(fn);
    } else {
        // We must save some data
        Ptr<Stream> pvm = dir.openFile(fn, Stream::C_CREATE);

        // Create header
        char header[30];
        storeUint32(&header[oSignature], MAGIC);
        std::memcpy(&header[oTimestamp],
                    getCurrentTurn().getCurrentUniverse().getTimestamp().getRawData(),
                    GTimestamp::SIZE);
        header[oEOF] = 26;
        header[oVersion] = VERSION;
        storeUint16(&header[oHeaderSize], 4);
        storeUint16(&header[oTurn], getCurrentTurn().getTurnNumber());
        storeUint16(&header[oPlayer], player);
        pvm->writeT(header, sizeof(header));

        // Write content
        IntVMSaveContext ctx;
        for (IntProcessList::const_iterator i = pl.begin(); i != pl.end(); ++i)
            ctx.addProcess(**i);
        ctx.save(*pvm);
    }
}
