/**
  *  \file reshack/picedit.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/picedit.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "io/dirs.h"
#include "reshack/painter.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "reshack/select.h"
#include "reshack/colors.h"
#include "ui/msgbox.h"
#include "reshack/colorsel.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/focusiterator.h"
#include "util/except.h"
#include "io/sdlstr.h"
#include "io/io.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"
#include "ui/widgets/radiobutton.h"
#include "ui/uiglobal.h"
#include "util/strlist.h"
#include "ui/listbox.h"
#include "ui/widgets/decimalsel.h"
#include "ui/util.h"

namespace {
    enum {
        ffGFX,
        ffCC,
        ffCD,
        ffBMP
    };

    class RHSaveImageDialog : public UIWindow {
        ObjectHolder h;
        Observable<int> file_format;
        UIInputLine file_name;
        void init();
        void onFileFormatChange(int n);
     public:
        RHSaveImageDialog();
        ~RHSaveImageDialog();

        void setFileName(string_t fn)
            { file_name.setText(fn); }
        string_t getFileName() const
            { return file_name.getText(); }
        void setFileFormat(int ffmt)
            { file_format.set(ffmt); }
        int getFileFormat() const
            { return file_format.get(); }
    };

    class RHPictureEditorWindow : public UIWindow {
        ObjectHolder h;
        RHPainter* painter;
        void init(Ptr<GfxPixmap> pixmap, ColorMode color_mode);
     public:
        RHPictureEditorWindow(Ptr<GfxPixmap> pixmap, ColorMode color_mode, string_t title);
        ~RHPictureEditorWindow();
        void onModify();
        Ptr<GfxPixmap> getPixmap()
            { return painter->getPixmap(); }
    };
}

static const char*const extensions[] = { ".gfx", ".cc", ".cd", ".bmp" };

static int
getFormatFromFileName(string_t fn)
{
    fn = Directory::getFileName(fn);
    string_t::size_type n = fn.rfind('.');
    if (n != fn.npos) {
        fn.erase(0, n);
        for (int i = 0; i < int(countof(extensions)); ++i)
            if (fn == extensions[i])
                return i;
    }
    return -1;
}

static void
savePixelData(GfxPixmap& pix, Stream& os, bool translate)
{
    std::vector<uint8_t> line(pix.getWidth());
    for (int y = 0; y < pix.getHeight(); ++y) {
        for (int x = 0; x < pix.getWidth(); ++x) {
            line[x] = pix.getPixel(GfxPoint(x, y));
        }
        if (translate) {
            for (int x = 0; x < pix.getWidth(); ++x) {
                if (line[x] >= 0x10) {
                    line[x] += 0x20;
                }
            }
        }
        os.writeT((const char*)&line[0], pix.getWidth());
    }
}

static void
savePixelData16(GfxPixmap& pix, Stream& os)
{
    for (int y = 0; y < pix.getHeight(); ++y) {
        std::vector<uint8_t> line((pix.getWidth()+1) / 2);
        for (int x = 0; x < pix.getWidth(); ++x) {
            uint8_t pixel = pix.getPixel(GfxPoint(x, y));
            if (pixel >= 0xA0)
                pixel = pixel - 0xA0 + 10;
            pixel &= 15;
            if (x & 1)
                pixel <<= 4;
            line[x/2] |= pixel;
        }
        
        os.writeT((const char*)&line[0], line.size());
    }
}

/*************************** RHSaveImageDialog ***************************/

RHSaveImageDialog::RHSaveImageDialog()
    : UIWindow(_("Save Image"), BLUE_WINDOW, COLOR_GRAY),
      file_format(-1),
      file_name(1000, 20, ss_Alt + 'n', UIInputLine::ilf_Framed, 1000)
{
    init();
}

RHSaveImageDialog::~RHSaveImageDialog()
{ }

void
RHSaveImageDialog::init()
{
    // VBox
    //   HBox
    //     "File Name"
    //     UIInputLine
    //   HBox
    //     "File Type"
    //     VBox
    //       UIRadioButton^n
    //   HBox
    //     UISpacer
    //     UIButton "OK"
    //     UIButton "Cancel"

    setLayout(UIVBoxLayout::instance5);
    UIGroup& g1  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g22 = h.add(new UIGroup(UIHBoxLayout::instance5, 1001));
    UIGroup& g3  = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIStaticText(_("File Name:"), UIStaticText::StaticTitle)));
    g1.add(file_name);

    g2.add(h.add(new UIStaticText(_("File Format:"), UIStaticText::StaticTitle)));
    g2.add(g22);

    g22.add(h.add(new UIRadioButton(2000, _(".gfx"), ss_Alt + 'g', file_format, ffGFX)));
    g22.add(h.add(new UIRadioButton(2001, _(".cc"),  ss_Alt + 'c', file_format, ffCC)));
    g22.add(h.add(new UIRadioButton(2002, _(".cd"),  ss_Alt + 'd', file_format, ffCD)));
    g22.add(h.add(new UIRadioButton(2003, _(".bmp"), ss_Alt + 'b', file_format, ffBMP)));

    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(2, _("Cancel"), SDLK_ESCAPE)).addStop());

    add(g1);
    add(g2);
    add(g3);
    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab)));
    add(h.add(new UIFocusIterator(2000, 2003, UIFocusIterator::fi_Horiz)));
    pack();

    file_format.sig_changed.add(this, &RHSaveImageDialog::onFileFormatChange);
}

void
RHSaveImageDialog::onFileFormatChange(int n)
{
    if (n >= 0 && n < int(countof(extensions))) {
        string_t dir = Directory::getDirName(file_name.getText());
        string_t fn  = Directory::getFileName(file_name.getText());
        string_t::size_type dot = fn.rfind('.');
        if (dot != fn.npos)
            fn.erase(dot);
        file_name.setText(Directory::makeFileName(dir, fn + extensions[n]));
    }
}

/************************* RHPictureEditorWindow *************************/

RHPictureEditorWindow::RHPictureEditorWindow(Ptr<GfxPixmap> pixmap, ColorMode color_mode, string_t title)
    : UIWindow(title, BLUE_WINDOW, COLOR_BLACK)
{
    init(pixmap, color_mode);
}

RHPictureEditorWindow::~RHPictureEditorWindow()
{ }

void
RHPictureEditorWindow::init(Ptr<GfxPixmap> pixmap, ColorMode color_mode)
{
    // HBox
    //   VBox
    //     RHToolSelector
    //     RHZoomSelector
    //     UISpacer
    //   RHPainter
    //   VBox
    //     RHColorSelector
    //     UISpacer
    //     UIButton "close"

    UIGroup& group1 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group3 = h.add(new UIGroup(UIVBoxLayout::instance5));

    painter = &h.add(new RHPainter(pixmap, color_mode));
    RHToolSelector& sel = h.add(new RHToolSelector(*painter));
    sel.addNewTool(SDLK_F1, new RHPencilTool());
    sel.addNewTool(SDLK_F2, new RHBrushTool());
    sel.addNewTool(SDLK_F3, new RHLineTool());
    sel.addNewTool(SDLK_F4, new RHRectangleTool());
    sel.addNewTool(SDLK_F5, new RHBarTool());
    sel.addNewTool(SDLK_F6, new RHCircleTool());
    sel.addNewTool(SDLK_F7, new RHCopyTool(&sel));
    sel.addNewTool(SDLK_F8, new RHPasteTool(true));
    sel.addNewTool(SDLK_F9, new RHPasteTool(false));
    sel.addNewTool(SDLK_F10, new RHPipetteTool());

    setLayout(UIHBoxLayout::instance5);
    add(group1);
    add(*painter);
    add(group3);

    group1.add(sel);
    group1.add(h.add(new RHZoomSelector(*painter, 4)));
    group1.add(h.add(new UISpacer()));

    group3.add(h.add(new RHColorSelector(*painter)));
    group3.add(h.add(new UISpacer()));
    group3.add(h.add(new UIButton(1, _("Modify"), 'm')).addCall(this, &RHPictureEditorWindow::onModify));
    group3.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());

    pack();
}

void
RHPictureEditorWindow::onModify()
{
    UIBaseWidget* w = getWidgetById(1);
    if (!w) {
        return;
    }

    StringList list;
    list.add(1, _("Set width"));
    list.add(2, _("Set height"));
    list.add(3, _("Add X aux line"));
    list.add(4, _("Add Y aux line"));
    list.add(5, _("Remove X aux line"));
    list.add(6, _("Remove Y aux line"));

    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    if (!box.doStandardMenu(*w)) {
        return;
    }

    UIDecimalSelector nsel(0, 0, 1024, 16);
    if (!nsel.doStandardDialog(list.get(box.getCurrent()), list.get(box.getCurrent()), string_t())) {
        return;
    }

    switch (box.getKey()) {
     case 1:
        painter->setSize(GfxPoint(nsel.getValue(), painter->getSize().y));
        break;
     case 2:
        painter->setSize(GfxPoint(painter->getSize().x, nsel.getValue()));
        break;
     case 3:
        painter->addAuxLine(nsel.getValue(), true);
        break;
     case 4:
        painter->addAuxLine(nsel.getValue(), false);
        break;
     case 5:
        painter->removeAuxLine(nsel.getValue(), true);
        break;
     case 6:
        painter->removeAuxLine(nsel.getValue(), false);
        break;
    }
}

/**************************** RHPictureEditor ****************************/

RHPictureEditor::RHPictureEditor(Ptr<GfxPixmap> pixmap, string_t file_name)
    : pixmap(pixmap),
      file_name(file_name),
      dirty(false),
      color_mode(identifyColorMode(pixmap))
{ }

RHPictureEditor::RHPictureEditor(Ptr<GfxPixmap> pixmap, ColorMode color_mode)
    : pixmap(pixmap),
      file_name(""),
      dirty(false),
      color_mode(color_mode)
{ }

RHPictureEditor::~RHPictureEditor()
{ }

string_t
RHPictureEditor::getName()
{
    string_t name;
    if (file_name.empty())
        name = _("Unnamed");
    else
        name = Directory::getFileName(file_name);
    name += format(" (%dx%d)", pixmap->getWidth(), pixmap->getHeight());
    return name;
}

void
RHPictureEditor::edit()
{
    // Can we edit this?
    if (ui_screen->getBitsPerPixel() <= 8) {
        if (color_mode != StandardPaletteColor && color_mode != GrayscaleColor) {
            messageBox(_("The screen color depth is not sufficient to edit this picture."),
                       _("Error"));
            return;
        }
    }

    // Make it editable
    pixmap = makeEditable(pixmap);
    setScreenPalette(pixmap);

    RHPictureEditorWindow win(pixmap, color_mode, Directory::getFileName(file_name));
    win.setExtent(ui_root->getExtent());
    ui_root->runModal(win);
    pixmap = win.getPixmap();
    dirty = true;
}

void
RHPictureEditor::save()
{
    // Create dialog
    RHSaveImageDialog dlg;
    dlg.setFileName(file_name);

    // Figure out file format
    int namefmt = getFormatFromFileName(file_name);
    ColorMode cm = identifyColorMode(pixmap);
    switch (cm) {
     case TrueColor:
        // Truecolor can be saved in BMP only
        dlg.setFileFormat(ffBMP);
        dlg.setStateOf(2000 + ffGFX, UIWindow::st_Disabled, true);
        dlg.setStateOf(2000 + ffCC,  UIWindow::st_Disabled, true);
        dlg.setStateOf(2000 + ffCD,  UIWindow::st_Disabled, true);
        break;
     case ForeignPaletteColor:
        // Those can be saved in BMP, or in palette-agnostic GFX
        if (namefmt == ffGFX)
            dlg.setFileFormat(ffGFX);
        else
            dlg.setFileFormat(ffBMP);
        dlg.setStateOf(2000 + ffCC, UIWindow::st_Disabled, true);
        dlg.setStateOf(2000 + ffCD, UIWindow::st_Disabled, true);
        break;

     case StandardPaletteColor:
     case GrayscaleColor:
        // Those can be saved in anything
        if (namefmt >= 0)
            dlg.setFileFormat(namefmt);
        else
            dlg.setFileFormat(ffBMP);
        break;
    }

    ui_root->centerWidget(dlg);

    if (ui_root->runModal(dlg) == 1) {
        // OK
        try {
            saveImage(dlg.getFileName(), dlg.getFileFormat());
            dirty = false;
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to save file %s: %s"), e.getFileName(), e.what()), _("Error"));
        }
        catch (std::exception& e) {
            messageBox(format(_("Unable to save file %s: %s"), dlg.getFileName(), e.what()), _("Error"));
        }
    }
}

void
RHPictureEditor::saveImage(string_t name, int format)
{
    Ptr<Stream> os = Stream::openFile(name, Stream::C_CREATE);
    switch (format) {
     case ffBMP:
        {
            StreamSDLInterface ifc(*os);
            SDL_SaveBMP_RW(pixmap->getSystemSurface(), &ifc, false);
        }
        break;

     case ffGFX:
        {
            // 0x800, width, height, 8-bpp raw pixel data
            char header[6];
            storeInt16(header,   0x800);
            storeInt16(header+2, pixmap->getWidth());
            storeInt16(header+4, pixmap->getHeight());
            os->writeT(header, sizeof(header));

            savePixelData(*pixmap, *os, false);
        }
        break;

     case ffCC:
        {
            // 'CC', width, height, 5 RGB triplets, 4-bpp pixel data in PCC 1.x 16-color palette
            char header[6 + 5*3];
            SDL_Palette* pal = pixmap->getSystemSurface()->format->palette;

            header[0] = 'C';
            header[1] = 'C';
            storeInt16(header+2, pixmap->getWidth());
            storeInt16(header+4, pixmap->getHeight());
            for (int i = 0; i < 5; ++i) {
                if (pal && pal->ncolors > 160+i) {
                    header[6 + 3*i]     = pal->colors[160+i].r / 4;
                    header[6 + 3*i + 1] = pal->colors[160+i].g / 4;
                    header[6 + 3*i + 2] = pal->colors[160+i].b / 4;
                } else {
                    header[6 + 3*i]     = 0;
                    header[6 + 3*i + 1] = 0;
                    header[6 + 3*i + 2] = 0;
                }
            }
            os->writeT(header, sizeof(header));

            savePixelData16(*pixmap, *os);
        }
        break;

     case ffCD:
        {
            // 'CD', width, height, 64 RGB triplets, 8-bpp pixel data in PCC 1.x 256-color palette
            char header[6 + 64*3];
            SDL_Palette* pal = pixmap->getSystemSurface()->format->palette;

            header[0] = 'C';
            header[1] = 'D';
            storeInt16(header+2, pixmap->getWidth());
            storeInt16(header+4, pixmap->getHeight());
            for (int i = 0; i < 64; ++i) {
                if (pal && pal->ncolors > 160+i) {
                    header[6 + 3*i]     = pal->colors[160+i].r / 4;
                    header[6 + 3*i + 1] = pal->colors[160+i].g / 4;
                    header[6 + 3*i + 2] = pal->colors[160+i].b / 4;
                } else {
                    header[6 + 3*i]     = 0;
                    header[6 + 3*i + 1] = 0;
                    header[6 + 3*i + 2] = 0;
                }
            }
            os->writeT(header, sizeof(header));

            savePixelData(*pixmap, *os, true);
        }
        break;
    }
}

bool
RHPictureEditor::hasUnsavedChanges()
{
    return dirty;
}
